text {
  font-family: monospace;
}
.node {
  stroke: #BDBDBD;
  stroke-width: 1px;
  fill: #e6ffff;
  border: 2px;
  /*.attr("fill", "url(#blue_grad)")*/
}

.link {
  stroke: #000;
  stroke-opacity: .6;
  fill: none;
}
.link-selected {
  stroke: orange;
  stroke-opacity: .9;
  fill: none;
}

.port {
  stroke: #000;
  opacity: .6;
}
tspan {
  white-space:pre;
}
.node-external-port {
  stroke: #000;
  stroke-width: 0px;
  fill: #BDBDBD;
  border: 1px;
}
body {
  margin: 0;
}